#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import hashlib
import json

import fixtures
from stevedore import extension

from heat.common import exception
from heat.common import template_format
from heat.engine.cfn import functions as cfn_funcs
from heat.engine.cfn import parameters as cfn_p
from heat.engine.cfn import template as cfn_t
from heat.engine.clients.os import nova
from heat.engine import environment
from heat.engine import function
from heat.engine.hot import template as hot_t
from heat.engine import node_data
from heat.engine import rsrc_defn
from heat.engine import stack
from heat.engine import stk_defn
from heat.engine import template
from heat.tests import common
from heat.tests.openstack.nova import fakes as fakes_nova
from heat.tests import utils

mapping_template = template_format.parse('''{
  "AWSTemplateFormatVersion" : "2010-09-09",
  "Mappings" : {
    "ValidMapping" : {
      "TestKey" : { "TestValue" : "wibble" }
    },
    "InvalidMapping" : {
      "ValueList" : [ "foo", "bar" ],
      "ValueString" : "baz"
    },
    "MapList": [ "foo", { "bar" : "baz" } ],
    "MapString": "foobar"
  }
}''')

empty_template = template_format.parse('''{
  "HeatTemplateFormatVersion" : "2012-12-12",
}''')

aws_empty_template = template_format.parse('''{
  "AWSTemplateFormatVersion" : "2010-09-09",
}''')

parameter_template = template_format.parse('''{
  "HeatTemplateFormatVersion" : "2012-12-12",
  "Parameters" : {
    "foo" : { "Type" : "String" },
    "blarg" : { "Type" : "String", "Default": "quux" }
  }
}''')


resource_template = template_format.parse('''{
  "HeatTemplateFormatVersion" : "2012-12-12",
  "Resources" : {
    "foo" : { "Type" : "GenericResourceType" },
    "blarg" : { "Type" : "GenericResourceType" }
  }
}''')


def join(raw):
    tmpl = template.Template(mapping_template)
    return function.resolve(tmpl.parse(None, raw))


class DummyClass(object):
    metadata = None

    def metadata_get(self):
        return self.metadata

    def metadata_set(self, metadata):
        self.metadata = metadata


class TemplatePluginFixture(fixtures.Fixture):
    def __init__(self, templates=None):
        templates = templates or {}
        super(TemplatePluginFixture, self).__init__()
        self.templates = [extension.Extension(k, None, v, None)
                          for (k, v) in templates.items()]

    def _get_template_extension_manager(self):
        return extension.ExtensionManager.make_test_instance(self.templates)

    def setUp(self):
        super(TemplatePluginFixture, self).setUp()

        def clear_template_classes():
            template._template_classes = None

        clear_template_classes()
        self.useFixture(fixtures.MockPatchObject(
            template,
            '_get_template_extension_manager',
            new=self._get_template_extension_manager))
        self.addCleanup(clear_template_classes)


class TestTemplatePluginManager(common.HeatTestCase):
    def test_template_NEW_good(self):
        class NewTemplate(template.Template):
            SECTIONS = (VERSION, MAPPINGS, CONDITIONS, PARAMETERS) = (
                'NEWTemplateFormatVersion',
                '__undefined__',
                'conditions',
                'parameters')
            RESOURCES = 'thingies'

            def param_schemata(self, param_defaults=None):
                pass

            def get_section_name(self, section):
                pass

            def parameters(self, stack_identifier, user_params,
                           param_defaults=None):
                pass

            def resource_definitions(self, stack):
                pass

            def add_resource(self, definition, name=None):
                pass

            def outputs(self, stack):
                pass

            def __getitem__(self, section):
                return {}

        class NewTemplatePrint(function.Function):
            def result(self):
                return 'always this'

        self.useFixture(TemplatePluginFixture(
            {'NEWTemplateFormatVersion.2345-01-01': NewTemplate}))

        t = {'NEWTemplateFormatVersion': '2345-01-01'}
        tmpl = template.Template(t)
        err = tmpl.validate()
        self.assertIsNone(err)


class TestTemplateVersion(common.HeatTestCase):

    versions = (('heat_template_version', '2013-05-23'),
                ('HeatTemplateFormatVersion', '2012-12-12'),
                ('AWSTemplateFormatVersion', '2010-09-09'))

    def test_hot_version(self):
        tmpl = {
            'heat_template_version': '2013-05-23',
            'foo': 'bar',
            'parameters': {}
        }
        self.assertEqual(('heat_template_version', '2013-05-23'),
                         template.get_version(tmpl, self.versions))

    def test_cfn_version(self):
        tmpl = {
            'AWSTemplateFormatVersion': '2010-09-09',
            'foo': 'bar',
            'Parameters': {}
        }
        self.assertEqual(('AWSTemplateFormatVersion', '2010-09-09'),
                         template.get_version(tmpl, self.versions))

    def test_heat_cfn_version(self):
        tmpl = {
            'HeatTemplateFormatVersion': '2012-12-12',
            'foo': 'bar',
            'Parameters': {}
        }
        self.assertEqual(('HeatTemplateFormatVersion', '2012-12-12'),
                         template.get_version(tmpl, self.versions))

    def test_missing_version(self):
        tmpl = {
            'foo': 'bar',
            'Parameters': {}
        }
        ex = self.assertRaises(exception.InvalidTemplateVersion,
                               template.get_version, tmpl, self.versions)
        self.assertEqual('The template version is invalid: Template version '
                         'was not provided', str(ex))

    def test_ambiguous_version(self):
        tmpl = {
            'AWSTemplateFormatVersion': '2010-09-09',
            'HeatTemplateFormatVersion': '2012-12-12',
            'foo': 'bar',
            'Parameters': {}
        }
        self.assertRaises(exception.InvalidTemplateVersion,
                          template.get_version, tmpl, self.versions)


class ParserTest(common.HeatTestCase):

    def test_list(self):
        raw = ['foo', 'bar', 'baz']
        parsed = join(raw)
        for i in range(len(raw)):
            self.assertEqual(raw[i], parsed[i])
        self.assertIsNot(raw, parsed)

    def test_dict(self):
        raw = {'foo': 'bar', 'blarg': 'wibble'}
        parsed = join(raw)
        for k in raw:
            self.assertEqual(raw[k], parsed[k])
        self.assertIsNot(raw, parsed)

    def test_dict_list(self):
        raw = {'foo': ['bar', 'baz'], 'blarg': 'wibble'}
        parsed = join(raw)
        self.assertEqual(raw['blarg'], parsed['blarg'])
        for i in range(len(raw['foo'])):
            self.assertEqual(raw['foo'][i], parsed['foo'][i])
        self.assertIsNot(raw, parsed)
        self.assertIsNot(raw['foo'], parsed['foo'])

    def test_list_dict(self):
        raw = [{'foo': 'bar', 'blarg': 'wibble'}, 'baz', 'quux']
        parsed = join(raw)
        for i in range(1, len(raw)):
            self.assertEqual(raw[i], parsed[i])
        for k in raw[0]:
            self.assertEqual(raw[0][k], parsed[0][k])
        self.assertIsNot(raw, parsed)
        self.assertIsNot(raw[0], parsed[0])

    def test_join(self):
        raw = {'Fn::Join': [' ', ['foo', 'bar', 'baz']]}
        self.assertEqual('foo bar baz', join(raw))

    def test_join_none(self):
        raw = {'Fn::Join': [' ', ['foo', None, 'baz']]}
        self.assertEqual('foo  baz', join(raw))

    def test_join_list(self):
        raw = [{'Fn::Join': [' ', ['foo', 'bar', 'baz']]}, 'blarg', 'wibble']
        parsed = join(raw)
        self.assertEqual('foo bar baz', parsed[0])
        for i in range(1, len(raw)):
            self.assertEqual(raw[i], parsed[i])
        self.assertIsNot(raw, parsed)

    def test_join_dict_val(self):
        raw = {'quux': {'Fn::Join': [' ', ['foo', 'bar', 'baz']]},
               'blarg': 'wibble'}
        parsed = join(raw)
        self.assertEqual('foo bar baz', parsed['quux'])
        self.assertEqual(raw['blarg'], parsed['blarg'])
        self.assertIsNot(raw, parsed)


class TestTemplateConditionParser(common.HeatTestCase):

    def setUp(self):
        super(TestTemplateConditionParser, self).setUp()
        self.ctx = utils.dummy_context()
        t = {
            'heat_template_version': '2016-10-14',
            'parameters': {
                'env_type': {
                    'type': 'string',
                    'default': 'test'
                }
            },
            'conditions': {
                'prod_env': {
                    'equals': [{'get_param': 'env_type'}, 'prod']}},
            'resources': {
                'r1': {
                    'type': 'GenericResourceType',
                    'condition': 'prod_env'
                }
            },
            'outputs': {
                'foo': {
                    'condition': 'prod_env',
                    'value': 'show me'
                }
            }
        }

        self.tmpl = template.Template(t)

    def test_conditions_with_non_supported_functions(self):
        t = {
            'heat_template_version': '2016-10-14',
            'parameters': {
                'env_type': {
                    'type': 'string',
                    'default': 'test'
                }
            },
            'conditions': {
                'prod_env': {
                    'equals': [{'get_param': 'env_type'},
                               {'get_attr': [None, 'att']}]}}}
        # test with get_attr in equals
        tmpl = template.Template(t)
        stk = stack.Stack(self.ctx, 'test_condition_with_get_attr_func', tmpl)
        ex = self.assertRaises(exception.StackValidationFailed,
                               tmpl.conditions, stk)
        self.assertIn('"get_attr" is invalid', str(ex))
        self.assertIn('conditions.prod_env.equals[1].get_attr',
                      str(ex))

        # test with get_resource in top level of a condition
        tmpl.t['conditions']['prod_env'] = {'get_resource': 'R1'}
        stk = stack.Stack(self.ctx, 'test_condition_with_get_attr_func', tmpl)
        ex = self.assertRaises(exception.StackValidationFailed,
                               tmpl.conditions, stk)
        self.assertIn('"get_resource" is invalid', str(ex))

        # test with get_attr in top level of a condition
        tmpl.t['conditions']['prod_env'] = {'get_attr': [None, 'att']}
        stk = stack.Stack(self.ctx, 'test_condition_with_get_attr_func', tmpl)
        ex = self.assertRaises(exception.StackValidationFailed,
                               tmpl.conditions, stk)
        self.assertIn('"get_attr" is invalid', str(ex))

    def test_condition_resolved_not_boolean(self):
        t = {
            'heat_template_version': '2016-10-14',
            'parameters': {
                'env_type': {
                    'type': 'string',
                    'default': 'test'
                }
            },
            'conditions': {
                'prod_env': {'get_param': 'env_type'}}}

        # test with get_attr in equals
        tmpl = template.Template(t)
        stk = stack.Stack(self.ctx, 'test_condition_not_boolean', tmpl)

        conditions = tmpl.conditions(stk)
        ex = self.assertRaises(exception.StackValidationFailed,
                               conditions.is_enabled, 'prod_env')
        self.assertIn('The definition of condition "prod_env" is invalid',
                      str(ex))

    def test_condition_reference_condition(self):
        t = {
            'heat_template_version': '2016-10-14',
            'parameters': {
                'env_type': {
                    'type': 'string',
                    'default': 'test'
                }
            },
            'conditions': {
                'prod_env': {'equals': [{'get_param': 'env_type'}, 'prod']},
                'test_env': {'not': 'prod_env'},
                'prod_or_test_env': {'or': ['prod_env', 'test_env']},
                'prod_and_test_env': {'and': ['prod_env', 'test_env']},
            }}

        # test with get_attr in equals
        tmpl = template.Template(t)
        stk = stack.Stack(self.ctx, 'test_condition_reference', tmpl)
        conditions = tmpl.conditions(stk)

        self.assertFalse(conditions.is_enabled('prod_env'))
        self.assertTrue(conditions.is_enabled('test_env'))
        self.assertTrue(conditions.is_enabled('prod_or_test_env'))
        self.assertFalse(conditions.is_enabled('prod_and_test_env'))

    def test_get_res_condition_invalid(self):
        tmpl = copy.deepcopy(self.tmpl)
        # test condition name is invalid
        stk = stack.Stack(self.ctx, 'test_res_invalid_condition', tmpl)

        conds = tmpl.conditions(stk)
        ex = self.assertRaises(ValueError, conds.is_enabled, 'invalid_cd')
        self.assertIn('Invalid condition "invalid_cd"', str(ex))
        # test condition name is not string
        ex = self.assertRaises(ValueError, conds.is_enabled, 111)
        self.assertIn('Invalid condition "111"', str(ex))

    def test_res_condition_using_boolean(self):
        tmpl = copy.deepcopy(self.tmpl)
        # test condition name is boolean
        stk = stack.Stack(self.ctx, 'test_res_cd_boolean', tmpl)

        conds = tmpl.conditions(stk)
        self.assertTrue(conds.is_enabled(True))
        self.assertFalse(conds.is_enabled(False))

    def test_parse_output_condition_invalid(self):
        stk = stack.Stack(self.ctx,
                          'test_output_invalid_condition',
                          self.tmpl)

        # test condition name is invalid
        self.tmpl.t['outputs']['foo']['condition'] = 'invalid_cd'
        ex = self.assertRaises(exception.StackValidationFailed,
                               lambda: stk.outputs)
        self.assertIn('Invalid condition "invalid_cd"', str(ex))
        self.assertIn('outputs.foo.condition', str(ex))
        # test condition name is not string
        self.tmpl.t['outputs']['foo']['condition'] = 222
        ex = self.assertRaises(exception.StackValidationFailed,
                               lambda: stk.outputs)
        self.assertIn('Invalid condition "222"', str(ex))
        self.assertIn('outputs.foo.condition', str(ex))

    def test_conditions_circular_ref(self):
        t = {
            'heat_template_version': '2016-10-14',
            'parameters': {
                'env_type': {
                    'type': 'string',
                    'default': 'test'
                }
            },
            'conditions': {
                'first_cond': {'not': 'second_cond'},
                'second_cond': {'not': 'third_cond'},
                'third_cond': {'not': 'first_cond'},
            }
        }
        tmpl = template.Template(t)
        stk = stack.Stack(self.ctx, 'test_condition_circular_ref', tmpl)
        conds = tmpl.conditions(stk)
        ex = self.assertRaises(exception.StackValidationFailed,
                               conds.is_enabled, 'first_cond')
        self.assertIn('Circular definition for condition "first_cond"',
                      str(ex))

    def test_parse_output_condition_boolean(self):
        t = copy.deepcopy(self.tmpl.t)
        t['outputs']['foo']['condition'] = True
        stk = stack.Stack(self.ctx,
                          'test_output_cd_boolean',
                          template.Template(t))

        self.assertEqual('show me', stk.outputs['foo'].get_value())

        t = copy.deepcopy(self.tmpl.t)
        t['outputs']['foo']['condition'] = False
        stk = stack.Stack(self.ctx,
                          'test_output_cd_boolean',
                          template.Template(t))
        self.assertIsNone(stk.outputs['foo'].get_value())

    def test_parse_output_condition_function(self):
        t = copy.deepcopy(self.tmpl.t)
        t['outputs']['foo']['condition'] = {'not': 'prod_env'}
        stk = stack.Stack(self.ctx,
                          'test_output_cd_function',
                          template.Template(t))

        self.assertEqual('show me', stk.outputs['foo'].get_value())


class TestTemplateValidate(common.HeatTestCase):

    def test_template_validate_cfn_check_t_digest(self):
        t = {
            'AWSTemplateFormatVersion': '2010-09-09',
            'Description': 'foo',
            'Parameters': {},
            'Mappings': {},
            'Resources': {
                'server': {
                    'Type': 'OS::Nova::Server'
                }
            },
            'Outputs': {},
        }

        tmpl = template.Template(t)
        self.assertIsNone(tmpl.t_digest)
        tmpl.validate()
        self.assertEqual(
            hashlib.sha256(str(t).encode('utf-8')).hexdigest(),
            tmpl.t_digest, 'invalid template digest')

    def test_template_validate_cfn_good(self):
        t = {
            'AWSTemplateFormatVersion': '2010-09-09',
            'Description': 'foo',
            'Parameters': {},
            'Mappings': {},
            'Resources': {
                'server': {
                    'Type': 'OS::Nova::Server'
                }
            },
            'Outputs': {},
        }

        tmpl = template.Template(t)
        err = tmpl.validate()
        self.assertIsNone(err)

        # test with alternate version key
        t = {
            'HeatTemplateFormatVersion': '2012-12-12',
            'Description': 'foo',
            'Parameters': {},
            'Mappings': {},
            'Resources': {
                'server': {
                    'Type': 'OS::Nova::Server'
                }
            },
            'Outputs': {},
        }

        tmpl = template.Template(t)
        err = tmpl.validate()
        self.assertIsNone(err)

    def test_template_validate_cfn_bad_section(self):
        t = {
            'AWSTemplateFormatVersion': '2010-09-09',
            'Description': 'foo',
            'Parameteers': {},
            'Mappings': {},
            'Resources': {
                'server': {
                    'Type': 'OS::Nova::Server'
                }
            },
            'Outputs': {},
        }

        tmpl = template.Template(t)
        err = self.assertRaises(exception.InvalidTemplateSection,
                                tmpl.validate)
        self.assertIn('Parameteers', str(err))

    def test_template_validate_cfn_empty(self):
        t = template_format.parse('''
            AWSTemplateFormatVersion: 2010-09-09
            Parameters:
            Resources:
            Outputs:
            ''')
        tmpl = template.Template(t)
        err = tmpl.validate()
        self.assertIsNone(err)

    def test_get_resources_good(self):
        """Test get resources successful."""

        t = template_format.parse('''
            AWSTemplateFormatVersion: 2010-09-09
            Resources:
              resource1:
                Type: AWS::EC2::Instance
                Properties:
                  property1: value1
                Metadata:
                  foo: bar
                DependsOn: dummy
                DeletionPolicy: dummy
                UpdatePolicy:
                  foo: bar
        ''')

        expected = {'resource1': {'Type': 'AWS::EC2::Instance',
                                  'Properties': {'property1': 'value1'},
                                  'Metadata': {'foo': 'bar'},
                                  'DependsOn': 'dummy',
                                  'DeletionPolicy': 'dummy',
                                  'UpdatePolicy': {'foo': 'bar'}}}

        tmpl = template.Template(t)
        self.assertEqual(expected, tmpl[tmpl.RESOURCES])

    def test_get_resources_bad_no_data(self):
        """Test get resources without any mapping."""

        t = template_format.parse('''
            AWSTemplateFormatVersion: 2010-09-09
            Resources:
              resource1:
        ''')

        tmpl = template.Template(t)
        error = self.assertRaises(exception.StackValidationFailed,
                                  tmpl.validate)
        self.assertEqual('Each Resource must contain a Type key.',
                         str(error))

    def test_get_resources_no_type(self):
        """Test get resources with invalid key."""

        t = template_format.parse('''
            AWSTemplateFormatVersion: 2010-09-09
            Resources:
              resource1:
                Properties:
                  property1: value1
                Metadata:
                  foo: bar
                DependsOn: dummy
                DeletionPolicy: dummy
                UpdatePolicy:
                  foo: bar
        ''')

        tmpl = template.Template(t)
        error = self.assertRaises(exception.StackValidationFailed,
                                  tmpl.validate)
        self.assertEqual('Each Resource must contain a Type key.',
                         str(error))

    def test_template_validate_hot_check_t_digest(self):
        t = {
            'heat_template_version': '2015-04-30',
            'description': 'foo',
            'parameters': {},
            'resources': {
                'server': {
                    'type': 'OS::Nova::Server'
                }
            },
            'outputs': {},
        }

        tmpl = template.Template(t)
        self.assertIsNone(tmpl.t_digest)
        tmpl.validate()
        self.assertEqual(hashlib.sha256(
            str(t).encode('utf-8')).hexdigest(),
            tmpl.t_digest, 'invalid template digest')

    def test_template_validate_hot_good(self):
        t = {
            'heat_template_version': '2013-05-23',
            'description': 'foo',
            'parameters': {},
            'resources': {
                'server': {
                    'type': 'OS::Nova::Server'
                }
            },
            'outputs': {},
        }

        tmpl = template.Template(t)
        err = tmpl.validate()
        self.assertIsNone(err)

    def test_template_validate_hot_bad_section(self):
        t = {
            'heat_template_version': '2013-05-23',
            'description': 'foo',
            'parameteers': {},
            'resources': {
                'server': {
                    'type': 'OS::Nova::Server'
                }
            },
            'outputs': {},
        }

        tmpl = template.Template(t)
        err = self.assertRaises(exception.InvalidTemplateSection,
                                tmpl.validate)
        self.assertIn('parameteers', str(err))


class TemplateTest(common.HeatTestCase):

    def setUp(self):
        super(TemplateTest, self).setUp()
        self.ctx = utils.dummy_context()

    @staticmethod
    def resolve(snippet, template, stack=None):
        return function.resolve(template.parse(stack and stack.defn, snippet))

    @staticmethod
    def resolve_condition(snippet, template, stack=None):
        return function.resolve(template.parse_condition(stack and stack.defn,
                                                         snippet))

    def test_defaults(self):
        empty = template.Template(empty_template)
        self.assertNotIn('AWSTemplateFormatVersion', empty)
        self.assertEqual('No description', empty['Description'])
        self.assertEqual({}, empty['Mappings'])
        self.assertEqual({}, empty['Resources'])
        self.assertEqual({}, empty['Outputs'])

    def test_aws_version(self):
        tmpl = template.Template(mapping_template)
        self.assertEqual(('AWSTemplateFormatVersion', '2010-09-09'),
                         tmpl.version)

    def test_heat_version(self):
        tmpl = template.Template(resource_template)
        self.assertEqual(('HeatTemplateFormatVersion', '2012-12-12'),
                         tmpl.version)

    def test_invalid_hot_version(self):
        invalid_hot_version_tmp = template_format.parse(
            '''{
            "heat_template_version" : "2012-12-12",
            }''')
        init_ex = self.assertRaises(exception.InvalidTemplateVersion,
                                    template.Template, invalid_hot_version_tmp)
        valid_versions = ['2013-05-23', '2014-10-16',
                          '2015-04-30', '2015-10-15', '2016-04-08',
                          '2016-10-14', '2017-02-24', '2017-09-01',
                          '2018-03-02', '2018-08-31', '2021-04-16',
                          'newton', 'ocata', 'pike',
                          'queens', 'rocky', 'wallaby']
        ex_error_msg = ('The template version is invalid: '
                        '"heat_template_version: 2012-12-12". '
                        '"heat_template_version" should be one of: %s'
                        % ', '.join(valid_versions))
        self.assertEqual(ex_error_msg, str(init_ex))

    def test_invalid_version_not_in_hot_versions(self):
        invalid_hot_version_tmp = template_format.parse(
            '''{
            "heat_template_version" : "2012-12-12",
            }''')
        versions = {
            ('heat_template_version', '2013-05-23'): hot_t.HOTemplate20130523,
            ('heat_template_version', '2013-06-23'): hot_t.HOTemplate20130523
        }

        temp_copy = copy.deepcopy(template._template_classes)
        template._template_classes = versions
        init_ex = self.assertRaises(exception.InvalidTemplateVersion,
                                    template.Template, invalid_hot_version_tmp)
        ex_error_msg = ('The template version is invalid: '
                        '"heat_template_version: 2012-12-12". '
                        '"heat_template_version" should be '
                        'one of: 2013-05-23, 2013-06-23')
        self.assertEqual(ex_error_msg, str(init_ex))
        template._template_classes = temp_copy

    def test_invalid_aws_version(self):
        invalid_aws_version_tmp = template_format.parse(
            '''{
            "AWSTemplateFormatVersion" : "2012-12-12",
            }''')
        init_ex = self.assertRaises(exception.InvalidTemplateVersion,
                                    template.Template, invalid_aws_version_tmp)
        ex_error_msg = ('The template version is invalid: '
                        '"AWSTemplateFormatVersion: 2012-12-12". '
                        '"AWSTemplateFormatVersion" should be: 2010-09-09')
        self.assertEqual(ex_error_msg, str(init_ex))

    def test_invalid_version_not_in_aws_versions(self):
        invalid_aws_version_tmp = template_format.parse(
            '''{
            "AWSTemplateFormatVersion" : "2012-12-12",
            }''')
        versions = {
            ('AWSTemplateFormatVersion', '2010-09-09'): cfn_t.CfnTemplate,
            ('AWSTemplateFormatVersion', '2011-06-23'): cfn_t.CfnTemplate
        }
        temp_copy = copy.deepcopy(template._template_classes)
        template._template_classes = versions

        init_ex = self.assertRaises(exception.InvalidTemplateVersion,
                                    template.Template, invalid_aws_version_tmp)
        ex_error_msg = ('The template version is invalid: '
                        '"AWSTemplateFormatVersion: 2012-12-12". '
                        '"AWSTemplateFormatVersion" should be '
                        'one of: 2010-09-09, 2011-06-23')
        self.assertEqual(ex_error_msg, str(init_ex))
        template._template_classes = temp_copy

    def test_invalid_heat_version(self):
        invalid_heat_version_tmp = template_format.parse(
            '''{
            "HeatTemplateFormatVersion" : "2010-09-09",
            }''')
        init_ex = self.assertRaises(exception.InvalidTemplateVersion,
                                    template.Template,
                                    invalid_heat_version_tmp)
        ex_error_msg = ('The template version is invalid: '
                        '"HeatTemplateFormatVersion: 2010-09-09". '
                        '"HeatTemplateFormatVersion" should be: 2012-12-12')
        self.assertEqual(ex_error_msg, str(init_ex))

    def test_invalid_version_not_in_heat_versions(self):
        invalid_heat_version_tmp = template_format.parse(
            '''{
            "HeatTemplateFormatVersion" : "2010-09-09",
            }''')
        versions = {
            ('HeatTemplateFormatVersion', '2012-12-12'): cfn_t.CfnTemplate,
            ('HeatTemplateFormatVersion', '2014-12-12'): cfn_t.CfnTemplate
        }
        temp_copy = copy.deepcopy(template._template_classes)
        template._template_classes = versions

        init_ex = self.assertRaises(exception.InvalidTemplateVersion,
                                    template.Template,
                                    invalid_heat_version_tmp)
        ex_error_msg = ('The template version is invalid: '
                        '"HeatTemplateFormatVersion: 2010-09-09". '
                        '"HeatTemplateFormatVersion" should be '
                        'one of: 2012-12-12, 2014-12-12')
        self.assertEqual(ex_error_msg, str(init_ex))

        template._template_classes = temp_copy

    def test_invalid_template(self):
        scanner_error = '''
            1
            Mappings:
              ValidMapping:
                TestKey: TestValue
            '''
        parser_error = '''
            Mappings:
              ValidMapping:
                TestKey: {TestKey1: "Value1" TestKey2: "Value2"}
            '''

        self.assertRaises(ValueError, template_format.parse, scanner_error)
        self.assertRaises(ValueError, template_format.parse, parser_error)

    def test_invalid_section(self):
        tmpl = template.Template({'HeatTemplateFormatVersion': '2012-12-12',
                                  'Foo': ['Bar']})
        self.assertNotIn('Foo', tmpl)

    def test_find_in_map(self):
        tmpl = template.Template(mapping_template)
        stk = stack.Stack(self.ctx, 'test', tmpl)
        find = {'Fn::FindInMap': ["ValidMapping", "TestKey", "TestValue"]}
        self.assertEqual("wibble", self.resolve(find, tmpl, stk))

    def test_find_in_invalid_map(self):
        tmpl = template.Template(mapping_template)
        stk = stack.Stack(self.ctx, 'test', tmpl)
        finds = ({'Fn::FindInMap': ["InvalidMapping", "ValueList", "foo"]},
                 {'Fn::FindInMap': ["InvalidMapping", "ValueString", "baz"]},
                 {'Fn::FindInMap': ["MapList", "foo", "bar"]},
                 {'Fn::FindInMap': ["MapString", "foo", "bar"]})

        for find in finds:
            self.assertRaises((KeyError, TypeError), self.resolve,
                              find, tmpl, stk)

    def test_bad_find_in_map(self):
        tmpl = template.Template(mapping_template)
        stk = stack.Stack(self.ctx, 'test', tmpl)
        finds = ({'Fn::FindInMap': "String"},
                 {'Fn::FindInMap': {"Dict": "String"}},
                 {'Fn::FindInMap': ["ShortList", "foo"]},
                 {'Fn::FindInMap': ["ReallyShortList"]})

        for find in finds:
            self.assertRaises(exception.StackValidationFailed,
                              self.resolve, find, tmpl, stk)

    def test_param_refs(self):
        env = environment.Environment({'foo': 'bar', 'blarg': 'wibble'})
        tmpl = template.Template(parameter_template, env=env)
        stk = stack.Stack(self.ctx, 'test', tmpl)
        p_snippet = {"Ref": "foo"}
        self.assertEqual("bar", self.resolve(p_snippet, tmpl, stk))

    def test_param_ref_missing(self):
        env = environment.Environment({'foo': 'bar'})
        tmpl = template.Template(parameter_template, env=env)
        stk = stack.Stack(self.ctx, 'test', tmpl)
        tmpl.env = environment.Environment({})
        stk.defn.parameters = cfn_p.CfnParameters(stk.identifier(), tmpl)
        snippet = {"Ref": "foo"}
        self.assertRaises(exception.UserParameterMissing,
                          self.resolve,
                          snippet, tmpl, stk)

    def test_resource_refs(self):
        tmpl = template.Template(resource_template)
        stk = stack.Stack(self.ctx, 'test', tmpl)
        stk.validate()

        data = node_data.NodeData.from_dict({'reference_id': 'bar'})
        stk_defn.update_resource_data(stk.defn, 'foo', data)
        r_snippet = {"Ref": "foo"}
        self.assertEqual("bar", self.resolve(r_snippet, tmpl, stk))

    def test_resource_refs_param(self):
        tmpl = template.Template(resource_template)
        stk = stack.Stack(self.ctx, 'test', tmpl)

        p_snippet = {"Ref": "baz"}
        parsed = tmpl.parse(stk.defn, p_snippet)
        self.assertIsInstance(parsed, cfn_funcs.ParamRef)

    def test_select_from_list(self):
        tmpl = template.Template(empty_template)
        data = {"Fn::Select": ["1", ["foo", "bar"]]}
        self.assertEqual("bar", self.resolve(data, tmpl))

    def test_select_from_list_integer_index(self):
        tmpl = template.Template(empty_template)
        data = {"Fn::Select": [1, ["foo", "bar"]]}
        self.assertEqual("bar", self.resolve(data, tmpl))

    def test_select_from_list_out_of_bound(self):
        tmpl = template.Template(empty_template)
        data = {"Fn::Select": ["0", ["foo", "bar"]]}
        self.assertEqual("foo", self.resolve(data, tmpl))
        data = {"Fn::Select": ["1", ["foo", "bar"]]}
        self.assertEqual("bar", self.resolve(data, tmpl))
        data = {"Fn::Select": ["2", ["foo", "bar"]]}
        self.assertEqual("", self.resolve(data, tmpl))

    def test_select_from_dict(self):
        tmpl = template.Template(empty_template)
        data = {"Fn::Select": ["red", {"red": "robin", "re": "foo"}]}
        self.assertEqual("robin", self.resolve(data, tmpl))

    def test_select_int_from_dict(self):
        tmpl = template.Template(empty_template)
        data = {"Fn::Select": ["2", {"1": "bar", "2": "foo"}]}
        self.assertEqual("foo", self.resolve(data, tmpl))

    def test_select_from_none(self):
        tmpl = template.Template(empty_template)
        data = {"Fn::Select": ["red", None]}
        self.assertEqual("", self.resolve(data, tmpl))

    def test_select_from_dict_not_existing(self):
        tmpl = template.Template(empty_template)
        data = {"Fn::Select": ["green", {"red": "robin", "re": "foo"}]}
        self.assertEqual("", self.resolve(data, tmpl))

    def test_select_from_serialized_json_map(self):
        tmpl = template.Template(empty_template)
        js = json.dumps({"red": "robin", "re": "foo"})
        data = {"Fn::Select": ["re", js]}
        self.assertEqual("foo", self.resolve(data, tmpl))

    def test_select_from_serialized_json_list(self):
        tmpl = template.Template(empty_template)
        js = json.dumps(["foo", "fee", "fum"])
        data = {"Fn::Select": ["0", js]}
        self.assertEqual("foo", self.resolve(data, tmpl))

    def test_select_empty_string(self):
        tmpl = template.Template(empty_template)
        data = {"Fn::Select": ["0", '']}
        self.assertEqual("", self.resolve(data, tmpl))
        data = {"Fn::Select": ["1", '']}
        self.assertEqual("", self.resolve(data, tmpl))
        data = {"Fn::Select": ["one", '']}
        self.assertEqual("", self.resolve(data, tmpl))

    def test_equals(self):
        tpl = template_format.parse('''
        AWSTemplateFormatVersion: 2010-09-09
        Parameters:
          env_type:
            Type: String
            Default: 'test'
        ''')
        snippet = {'Fn::Equals': [{'Ref': 'env_type'}, 'prod']}
        # when param 'env_type' is 'test', equals function resolve to false
        tmpl = template.Template(tpl)
        stk = stack.Stack(utils.dummy_context(),
                          'test_equals_false', tmpl)
        resolved = self.resolve_condition(snippet, tmpl, stk)
        self.assertFalse(resolved)
        # when param 'env_type' is 'prod', equals function resolve to true
        tmpl = template.Template(tpl,
                                 env=environment.Environment(
                                     {'env_type': 'prod'}))
        stk = stack.Stack(utils.dummy_context(),
                          'test_equals_true', tmpl)
        resolved = self.resolve_condition(snippet, tmpl, stk)
        self.assertTrue(resolved)

    def test_equals_invalid_args(self):
        tmpl = template.Template(aws_empty_template)

        snippet = {'Fn::Equals': ['test', 'prod', 'invalid']}
        exc = self.assertRaises(exception.StackValidationFailed,
                                self.resolve_condition, snippet, tmpl)

        error_msg = ('.Fn::Equals: Arguments to "Fn::Equals" must be '
                     'of the form: [value_1, value_2]')
        self.assertIn(error_msg, str(exc))
        # test invalid type
        snippet = {'Fn::Equals': {"equal": False}}
        exc = self.assertRaises(exception.StackValidationFailed,
                                self.resolve_condition, snippet, tmpl)
        self.assertIn(error_msg, str(exc))

    def test_not(self):
        tpl = template_format.parse('''
        AWSTemplateFormatVersion: 2010-09-09
        Parameters:
          env_type:
            Type: String
            Default: 'test'
        ''')
        snippet = {'Fn::Not': [{'Fn::Equals': [{'Ref': 'env_type'}, 'prod']}]}
        # when param 'env_type' is 'test', not function resolve to true
        tmpl = template.Template(tpl)
        stk = stack.Stack(utils.dummy_context(),
                          'test_not_true', tmpl)
        resolved = self.resolve_condition(snippet, tmpl, stk)
        self.assertTrue(resolved)
        # when param 'env_type' is 'prod', not function resolve to false
        tmpl = template.Template(tpl,
                                 env=environment.Environment(
                                     {'env_type': 'prod'}))
        stk = stack.Stack(utils.dummy_context(),
                          'test_not_false', tmpl)
        resolved = self.resolve_condition(snippet, tmpl, stk)
        self.assertFalse(resolved)

    def test_not_invalid_args(self):
        tmpl = template.Template(aws_empty_template)

        stk = stack.Stack(utils.dummy_context(),
                          'test_not_invalid', tmpl)
        snippet = {'Fn::Not': ['invalid_arg']}
        exc = self.assertRaises(ValueError,
                                self.resolve_condition, snippet, tmpl, stk)

        error_msg = 'Invalid condition "invalid_arg"'
        self.assertIn(error_msg, str(exc))
        # test invalid type
        snippet = {'Fn::Not': 'invalid'}
        exc = self.assertRaises(exception.StackValidationFailed,
                                self.resolve_condition, snippet, tmpl)
        error_msg = 'Arguments to "Fn::Not" must be '
        self.assertIn(error_msg, str(exc))

        snippet = {'Fn::Not': ['cd1', 'cd2']}
        exc = self.assertRaises(exception.StackValidationFailed,
                                self.resolve_condition, snippet, tmpl)
        error_msg = 'Arguments to "Fn::Not" must be '
        self.assertIn(error_msg, str(exc))

    def test_and(self):
        tpl = template_format.parse('''
        AWSTemplateFormatVersion: 2010-09-09
        Parameters:
          env_type:
            Type: String
            Default: 'test'
          zone:
            Type: String
            Default: 'shanghai'
        ''')
        snippet = {
            'Fn::And': [
                {'Fn::Equals': [{'Ref': 'env_type'}, 'prod']},
                {'Fn::Not': [{'Fn::Equals': [{'Ref': 'zone'}, "beijing"]}]}]}
        # when param 'env_type' is 'test', and function resolve to false
        tmpl = template.Template(tpl)
        stk = stack.Stack(utils.dummy_context(),
                          'test_and_false', tmpl)
        resolved = self.resolve_condition(snippet, tmpl, stk)
        self.assertFalse(resolved)
        # when param 'env_type' is 'prod', and param 'zone' is 'shanghai',
        # the 'and' function resolve to true
        tmpl = template.Template(tpl,
                                 env=environment.Environment(
                                     {'env_type': 'prod'}))
        stk = stack.Stack(utils.dummy_context(),
                          'test_and_true', tmpl)
        resolved = self.resolve_condition(snippet, tmpl, stk)
        self.assertTrue(resolved)
        # when param 'env_type' is 'prod', and param 'zone' is 'shanghai',
        # the 'and' function resolve to true
        tmpl = template.Template(tpl,
                                 env=environment.Environment(
                                     {'env_type': 'prod',
                                      'zone': 'beijing'}))
        stk = stack.Stack(utils.dummy_context(),
                          'test_and_false', tmpl)
        resolved = self.resolve_condition(snippet, tmpl, stk)
        self.assertFalse(resolved)

    def test_and_invalid_args(self):
        tmpl = template.Template(aws_empty_template)

        error_msg = ('The minimum number of condition arguments to "Fn::And" '
                     'is 2.')
        snippet = {'Fn::And': ['invalid_arg']}
        exc = self.assertRaises(exception.StackValidationFailed,
                                self.resolve_condition, snippet, tmpl)
        self.assertIn(error_msg, str(exc))

        error_msg = 'Arguments to "Fn::And" must be'
        # test invalid type
        snippet = {'Fn::And': 'invalid'}
        exc = self.assertRaises(exception.StackValidationFailed,
                                self.resolve_condition, snippet, tmpl)
        self.assertIn(error_msg, str(exc))

        stk = stack.Stack(utils.dummy_context(), 'test_and_invalid', tmpl)
        snippet = {'Fn::And': ['cd1', True]}
        exc = self.assertRaises(ValueError,
                                self.resolve_condition, snippet, tmpl, stk)
        error_msg = 'Invalid condition "cd1"'
        self.assertIn(error_msg, str(exc))

    def test_or(self):
        tpl = template_format.parse('''
        AWSTemplateFormatVersion: 2010-09-09
        Parameters:
          zone:
            Type: String
            Default: 'guangzhou'
        ''')
        snippet = {
            'Fn::Or': [
                {'Fn::Equals': [{'Ref': 'zone'}, 'shanghai']},
                {'Fn::Equals': [{'Ref': 'zone'}, 'beijing']}]}
        # when param 'zone' is neither equal to 'shanghai' nor 'beijing',
        # the 'or' function resolve to false
        tmpl = template.Template(tpl)
        stk = stack.Stack(utils.dummy_context(),
                          'test_or_false', tmpl)
        resolved = self.resolve_condition(snippet, tmpl, stk)
        self.assertFalse(resolved)
        # when param 'zone' equals to 'shanghai' or 'beijing',
        # the 'or' function resolve to true
        tmpl = template.Template(tpl,
                                 env=environment.Environment(
                                     {'zone': 'beijing'}))
        stk = stack.Stack(utils.dummy_context(),
                          'test_or_true', tmpl)
        resolved = self.resolve_condition(snippet, tmpl, stk)
        self.assertTrue(resolved)

        tmpl = template.Template(tpl,
                                 env=environment.Environment(
                                     {'zone': 'shanghai'}))
        stk = stack.Stack(utils.dummy_context(),
                          'test_or_true', tmpl)
        resolved = self.resolve_condition(snippet, tmpl, stk)
        self.assertTrue(resolved)

    def test_or_invalid_args(self):
        tmpl = template.Template(aws_empty_template)

        error_msg = ('The minimum number of condition arguments to "Fn::Or" '
                     'is 2.')
        snippet = {'Fn::Or': ['invalid_arg']}
        exc = self.assertRaises(exception.StackValidationFailed,
                                self.resolve_condition, snippet, tmpl)
        self.assertIn(error_msg, str(exc))

        error_msg = 'Arguments to "Fn::Or" must be'
        # test invalid type
        snippet = {'Fn::Or': 'invalid'}
        exc = self.assertRaises(exception.StackValidationFailed,
                                self.resolve_condition, snippet, tmpl)
        self.assertIn(error_msg, str(exc))

        stk = stack.Stack(utils.dummy_context(), 'test_or_invalid', tmpl)
        snippet = {'Fn::Or': ['invalid_cd', True]}
        exc = self.assertRaises(ValueError,
                                self.resolve_condition, snippet, tmpl, stk)
        error_msg = 'Invalid condition "invalid_cd"'
        self.assertIn(error_msg, str(exc))

    def test_join(self):
        tmpl = template.Template(empty_template)
        join = {"Fn::Join": [" ", ["foo", "bar"]]}
        self.assertEqual("foo bar", self.resolve(join, tmpl))

    def test_split_ok(self):
        tmpl = template.Template(empty_template)
        data = {"Fn::Split": [";", "foo; bar; achoo"]}
        self.assertEqual(['foo', ' bar', ' achoo'], self.resolve(data, tmpl))

    def test_split_no_delim_in_str(self):
        tmpl = template.Template(empty_template)
        data = {"Fn::Split": [";", "foo, bar, achoo"]}
        self.assertEqual(['foo, bar, achoo'], self.resolve(data, tmpl))

    def test_base64(self):
        tmpl = template.Template(empty_template)
        snippet = {"Fn::Base64": "foobar"}
        # For now, the Base64 function just returns the original text, and
        # does not convert to base64 (see issue #133)
        self.assertEqual("foobar", self.resolve(snippet, tmpl))

    def test_get_azs(self):
        tmpl = template.Template(empty_template)
        snippet = {"Fn::GetAZs": ""}
        self.assertEqual(["nova"], self.resolve(snippet, tmpl))

    def test_get_azs_with_stack(self):
        tmpl = template.Template(empty_template)
        snippet = {"Fn::GetAZs": ""}
        stk = stack.Stack(self.ctx, 'test_stack',
                          template.Template(empty_template))
        fc = fakes_nova.FakeClient()
        self.patchobject(nova.NovaClientPlugin, 'client', return_value=fc)
        self.assertEqual(["nova1"], self.resolve(snippet, tmpl, stk))

    def test_replace_string_values(self):
        tmpl = template.Template(empty_template)
        snippet = {"Fn::Replace": [
            {'$var1': 'foo', '%var2%': 'bar'},
            '$var1 is %var2%'
        ]}
        self.assertEqual('foo is bar', self.resolve(snippet, tmpl))

    def test_replace_number_values(self):
        tmpl = template.Template(empty_template)
        snippet = {"Fn::Replace": [
            {'$var1': 1, '%var2%': 2},
            '$var1 is not %var2%'
        ]}
        self.assertEqual('1 is not 2', self.resolve(snippet, tmpl))

        snippet = {"Fn::Replace": [
            {'$var1': 1.3, '%var2%': 2.5},
            '$var1 is not %var2%'
        ]}
        self.assertEqual('1.3 is not 2.5', self.resolve(snippet, tmpl))

    def test_replace_none_values(self):
        tmpl = template.Template(empty_template)
        snippet = {"Fn::Replace": [
            {'$var1': None, '${var2}': None},
            '"$var1" is "${var2}"'
        ]}
        self.assertEqual('"" is ""', self.resolve(snippet, tmpl))

    def test_replace_missing_key(self):
        tmpl = template.Template(empty_template)
        snippet = {"Fn::Replace": [
            {'$var1': 'foo', 'var2': 'bar'},
            '"$var1" is "${var3}"'
        ]}
        self.assertEqual('"foo" is "${var3}"', self.resolve(snippet, tmpl))

    def test_replace_param_values(self):
        env = environment.Environment({'foo': 'wibble'})
        tmpl = template.Template(parameter_template, env=env)
        stk = stack.Stack(self.ctx, 'test_stack', tmpl)
        snippet = {"Fn::Replace": [
            {'$var1': {'Ref': 'foo'}, '%var2%': {'Ref': 'blarg'}},
            '$var1 is %var2%'
        ]}
        self.assertEqual('wibble is quux', self.resolve(snippet, tmpl, stk))

    def test_member_list2map_good(self):
        tmpl = template.Template(empty_template)
        snippet = {"Fn::MemberListToMap": [
            'Name', 'Value', ['.member.0.Name=metric',
                              '.member.0.Value=cpu',
                              '.member.1.Name=size',
                              '.member.1.Value=56']]}
        self.assertEqual({'metric': 'cpu', 'size': '56'},
                         self.resolve(snippet, tmpl))

    def test_member_list2map_good2(self):
        tmpl = template.Template(empty_template)
        snippet = {"Fn::MemberListToMap": [
            'Key', 'Value', ['.member.2.Key=metric',
                             '.member.2.Value=cpu',
                             '.member.5.Key=size',
                             '.member.5.Value=56']]}
        self.assertEqual({'metric': 'cpu', 'size': '56'},
                         self.resolve(snippet, tmpl))

    def test_resource_facade(self):
        metadata_snippet = {'Fn::ResourceFacade': 'Metadata'}
        deletion_policy_snippet = {'Fn::ResourceFacade': 'DeletionPolicy'}
        update_policy_snippet = {'Fn::ResourceFacade': 'UpdatePolicy'}

        parent_resource = DummyClass()
        parent_resource.metadata_set({"foo": "bar"})

        parent_resource.t = rsrc_defn.ResourceDefinition(
            'parent', 'SomeType',
            deletion_policy=rsrc_defn.ResourceDefinition.RETAIN,
            update_policy={"blarg": "wibble"})
        tmpl = copy.deepcopy(empty_template)
        tmpl['Resources'] = {'parent': {'Type': 'SomeType',
                                        'DeletionPolicy': 'Retain',
                                        'UpdatePolicy': {"blarg": "wibble"}}}
        parent_resource.stack = stack.Stack(self.ctx, 'toplevel_stack',
                                            template.Template(tmpl))
        parent_resource.stack._resources = {'parent': parent_resource}

        stk = stack.Stack(self.ctx, 'test_stack',
                          template.Template(empty_template),
                          parent_resource='parent', owner_id=45)
        stk.set_parent_stack(parent_resource.stack)
        self.assertEqual({"foo": "bar"},
                         self.resolve(metadata_snippet, stk.t, stk))
        self.assertEqual('Retain',
                         self.resolve(deletion_policy_snippet, stk.t, stk))
        self.assertEqual({"blarg": "wibble"},
                         self.resolve(update_policy_snippet, stk.t, stk))

    def test_resource_facade_function(self):
        deletion_policy_snippet = {'Fn::ResourceFacade': 'DeletionPolicy'}

        parent_resource = DummyClass()
        parent_resource.metadata_set({"foo": "bar"})
        del_policy = cfn_funcs.Join(None,
                                    'Fn::Join', ['eta', ['R', 'in']])
        parent_resource.t = rsrc_defn.ResourceDefinition(
            'parent', 'SomeType',
            deletion_policy=del_policy)
        tmpl = copy.deepcopy(empty_template)
        tmpl['Resources'] = {'parent': {'Type': 'SomeType',
                                        'DeletionPolicy': del_policy}}
        parent_resource.stack = stack.Stack(self.ctx, 'toplevel_stack',
                                            template.Template(tmpl))
        parent_resource.stack._resources = {'parent': parent_resource}

        stk = stack.Stack(self.ctx, 'test_stack',
                          template.Template(empty_template),
                          parent_resource='parent')
        stk.set_parent_stack(parent_resource.stack)
        self.assertEqual('Retain',
                         self.resolve(deletion_policy_snippet, stk.t, stk))

    def test_resource_facade_invalid_arg(self):
        snippet = {'Fn::ResourceFacade': 'wibble'}
        stk = stack.Stack(self.ctx, 'test_stack',
                          template.Template(empty_template))
        error = self.assertRaises(exception.StackValidationFailed,
                                  self.resolve, snippet, stk.t, stk)
        self.assertIn(next(iter(snippet)), str(error))

    def test_resource_facade_missing_deletion_policy(self):
        snippet = {'Fn::ResourceFacade': 'DeletionPolicy'}

        parent_resource = DummyClass()
        parent_resource.metadata_set({"foo": "bar"})
        parent_resource.t = rsrc_defn.ResourceDefinition('parent', 'SomeType')
        tmpl = copy.deepcopy(empty_template)
        tmpl['Resources'] = {'parent': {'Type': 'SomeType'}}

        parent_resource.stack = stack.Stack(self.ctx, 'toplevel_stack',
                                            template.Template(tmpl))
        parent_resource.stack._resources = {'parent': parent_resource}
        stk = stack.Stack(self.ctx, 'test_stack',
                          template.Template(empty_template),
                          parent_resource='parent', owner_id=78)
        stk.set_parent_stack(parent_resource.stack)
        self.assertEqual('Delete', self.resolve(snippet, stk.t, stk))

    def test_prevent_parameters_access(self):
        expected_description = "This can be accessed"
        tmpl = template.Template({
            'AWSTemplateFormatVersion': '2010-09-09',
            'Description': expected_description,
            'Parameters': {
                'foo': {'Type': 'String', 'Required': True}
            }
        })
        self.assertEqual(expected_description, tmpl['Description'])
        keyError = self.assertRaises(KeyError, tmpl.__getitem__, 'Parameters')
        self.assertIn("can not be accessed directly", str(keyError))

    def test_parameters_section_not_iterable(self):
        expected_description = "This can be accessed"
        tmpl = template.Template({
            'AWSTemplateFormatVersion': '2010-09-09',
            'Description': expected_description,
            'Parameters': {
                'foo': {'Type': 'String', 'Required': True}
            }
        })
        self.assertEqual(expected_description, tmpl['Description'])
        self.assertNotIn('Parameters', tmpl.keys())

    def test_add_resource(self):
        cfn_tpl = template_format.parse('''
        AWSTemplateFormatVersion: 2010-09-09
        Resources:
          resource1:
            Type: AWS::EC2::Instance
            Properties:
              property1: value1
            Metadata:
              foo: bar
            DependsOn: dummy
            DeletionPolicy: Retain
            UpdatePolicy:
              foo: bar
          resource2:
            Type: AWS::EC2::Instance
          resource3:
            Type: AWS::EC2::Instance
            DependsOn:
              - resource1
              - dummy
              - resource2
        ''')
        source = template.Template(cfn_tpl)
        empty = template.Template(copy.deepcopy(empty_template))
        stk = stack.Stack(self.ctx, 'test_stack', source)

        for rname, defn in sorted(source.resource_definitions(stk).items()):
            empty.add_resource(defn)

        expected = copy.deepcopy(cfn_tpl['Resources'])
        del expected['resource1']['DependsOn']
        expected['resource3']['DependsOn'] = ['resource1', 'resource2']
        self.assertEqual(expected, empty.t['Resources'])

    def test_add_output(self):
        cfn_tpl = template_format.parse('''
        AWSTemplateFormatVersion: 2010-09-09
        Outputs:
          output1:
            Description: An output
            Value: foo
        ''')
        source = template.Template(cfn_tpl)
        empty = template.Template(copy.deepcopy(empty_template))
        stk = stack.Stack(self.ctx, 'test_stack', source)

        for defn in source.outputs(stk).values():
            empty.add_output(defn)

        self.assertEqual(cfn_tpl['Outputs'], empty.t['Outputs'])

    def test_create_empty_template_default_version(self):
        empty_template = template.Template.create_empty_template()
        self.assertEqual(hot_t.HOTemplate20150430, empty_template.__class__)
        self.assertEqual({}, empty_template['parameter_groups'])
        self.assertEqual({}, empty_template['resources'])
        self.assertEqual({}, empty_template['outputs'])

    def test_create_empty_template_returns_correct_version(self):
        t = template_format.parse('''
            AWSTemplateFormatVersion: 2010-09-09
            Parameters:
            Resources:
            Outputs:
            ''')
        aws_tmpl = template.Template(t)
        empty_template = template.Template.create_empty_template(
            version=aws_tmpl.version)
        self.assertEqual(aws_tmpl.__class__, empty_template.__class__)
        self.assertEqual({}, empty_template['Mappings'])
        self.assertEqual({}, empty_template['Resources'])
        self.assertEqual({}, empty_template['Outputs'])

        t = template_format.parse('''
            HeatTemplateFormatVersion: 2012-12-12
            Parameters:
            Resources:
            Outputs:
            ''')
        heat_tmpl = template.Template(t)
        empty_template = template.Template.create_empty_template(
            version=heat_tmpl.version)
        self.assertEqual(heat_tmpl.__class__, empty_template.__class__)
        self.assertEqual({}, empty_template['Mappings'])
        self.assertEqual({}, empty_template['Resources'])
        self.assertEqual({}, empty_template['Outputs'])

        t = template_format.parse('''
            heat_template_version: 2015-04-30
            parameter_groups:
            resources:
            outputs:
            ''')
        hot_tmpl = template.Template(t)
        empty_template = template.Template.create_empty_template(
            version=hot_tmpl.version)
        self.assertEqual(hot_tmpl.__class__, empty_template.__class__)
        self.assertEqual({}, empty_template['parameter_groups'])
        self.assertEqual({}, empty_template['resources'])
        self.assertEqual({}, empty_template['outputs'])

    def test_create_empty_template_from_another_template(self):
        res_param_template = template_format.parse('''{
          "HeatTemplateFormatVersion" : "2012-12-12",
          "Parameters" : {
            "foo" : { "Type" : "String" },
            "blarg" : { "Type" : "String", "Default": "quux" }
          },
          "Resources" : {
            "foo" : { "Type" : "GenericResourceType" },
            "blarg" : { "Type" : "GenericResourceType" }
          }
        }''')

        env = environment.Environment({'foo': 'bar'})
        hot_tmpl = template.Template(res_param_template, env)
        empty_template = template.Template.create_empty_template(
            from_template=hot_tmpl)
        self.assertEqual({}, empty_template['Resources'])
        self.assertEqual(hot_tmpl.env, empty_template.env)


class TemplateFnErrorTest(common.HeatTestCase):
    scenarios = [
        ('select_from_list_not_int',
         dict(expect=TypeError,
              snippet={"Fn::Select": ["one", ["foo", "bar"]]})),
        ('select_from_dict_not_str',
         dict(expect=TypeError,
              snippet={"Fn::Select": [1, {"red": "robin", "re": "foo"}]})),
        ('select_from_serialized_json_wrong',
         dict(expect=ValueError,
              snippet={"Fn::Select": ["not", "no json"]})),
        ('select_wrong_num_args_1',
         dict(expect=exception.StackValidationFailed,
              snippet={"Fn::Select": []})),
        ('select_wrong_num_args_2',
         dict(expect=exception.StackValidationFailed,
              snippet={"Fn::Select": ["4"]})),
        ('select_wrong_num_args_3',
         dict(expect=exception.StackValidationFailed,
              snippet={"Fn::Select": ["foo", {"foo": "bar"}, ""]})),
        ('select_wrong_num_args_4',
         dict(expect=TypeError,
              snippet={'Fn::Select': [['f'], {'f': 'food'}]})),
        ('split_no_delim',
         dict(expect=exception.StackValidationFailed,
              snippet={"Fn::Split": ["foo, bar, achoo"]})),
        ('split_no_list',
         dict(expect=exception.StackValidationFailed,
              snippet={"Fn::Split": "foo, bar, achoo"})),
        ('base64_list',
         dict(expect=TypeError,
              snippet={"Fn::Base64": ["foobar"]})),
        ('base64_dict',
         dict(expect=TypeError,
              snippet={"Fn::Base64": {"foo": "bar"}})),
        ('replace_list_value',
         dict(expect=TypeError,
              snippet={"Fn::Replace": [
                  {'$var1': 'foo', '%var2%': ['bar']},
                  '$var1 is %var2%']})),
        ('replace_list_mapping',
         dict(expect=exception.StackValidationFailed,
              snippet={"Fn::Replace": [
                  ['var1', 'foo', 'var2', 'bar'],
                  '$var1 is ${var2}']})),
        ('replace_dict',
         dict(expect=exception.StackValidationFailed,
              snippet={"Fn::Replace": {}})),
        ('replace_missing_template',
         dict(expect=exception.StackValidationFailed,
              snippet={"Fn::Replace": [['var1', 'foo', 'var2', 'bar']]})),
        ('replace_none_template',
         dict(expect=exception.StackValidationFailed,
              snippet={"Fn::Replace": [['var2', 'bar'], None]})),
        ('replace_list_string',
         dict(expect=TypeError,
              snippet={"Fn::Replace": [
                  {'var1': 'foo', 'var2': 'bar'},
                  ['$var1 is ${var2}']]})),
        ('join_string',
         dict(expect=TypeError,
              snippet={"Fn::Join": [" ", "foo"]})),
        ('join_dict',
         dict(expect=TypeError,
              snippet={"Fn::Join": [" ", {"foo": "bar"}]})),
        ('join_wrong_num_args_1',
         dict(expect=exception.StackValidationFailed,
              snippet={"Fn::Join": []})),
        ('join_wrong_num_args_2',
         dict(expect=exception.StackValidationFailed,
              snippet={"Fn::Join": [" "]})),
        ('join_wrong_num_args_3',
         dict(expect=exception.StackValidationFailed,
              snippet={"Fn::Join": [" ", {"foo": "bar"}, ""]})),
        ('join_string_nodelim',
         dict(expect=exception.StackValidationFailed,
              snippet={"Fn::Join": "o"})),
        ('join_string_nodelim_1',
         dict(expect=exception.StackValidationFailed,
              snippet={"Fn::Join": "oh"})),
        ('join_string_nodelim_2',
         dict(expect=exception.StackValidationFailed,
              snippet={"Fn::Join": "ohh"})),
        ('join_dict_nodelim1',
         dict(expect=exception.StackValidationFailed,
              snippet={"Fn::Join": {"foo": "bar"}})),
        ('join_dict_nodelim2',
         dict(expect=exception.StackValidationFailed,
              snippet={"Fn::Join": {"foo": "bar", "blarg": "wibble"}})),
        ('join_dict_nodelim3',
         dict(expect=exception.StackValidationFailed,
              snippet={"Fn::Join": {"foo": "bar", "blarg": "wibble",
                                    "baz": "quux"}})),
        ('member_list2map_no_key_or_val',
         dict(expect=exception.StackValidationFailed,
              snippet={"Fn::MemberListToMap": [
                  'Key', ['.member.2.Key=metric',
                          '.member.2.Value=cpu',
                          '.member.5.Key=size',
                          '.member.5.Value=56']]})),
        ('member_list2map_no_list',
         dict(expect=exception.StackValidationFailed,
              snippet={"Fn::MemberListToMap": [
                  'Key', '.member.2.Key=metric']})),
        ('member_list2map_not_string',
         dict(expect=exception.StackValidationFailed,
              snippet={"Fn::MemberListToMap": [
                  'Name', ['Value'], ['.member.0.Name=metric',
                                      '.member.0.Value=cpu',
                                      '.member.1.Name=size',
                                      '.member.1.Value=56']]})),
    ]

    def test_bad_input(self):
        tmpl = template.Template(empty_template)

        def resolve(s):
            return TemplateTest.resolve(s, tmpl)

        error = self.assertRaises(self.expect,
                                  resolve,
                                  self.snippet)
        self.assertIn(next(iter(self.snippet)), str(error))


class ResolveDataTest(common.HeatTestCase):

    def setUp(self):
        super(ResolveDataTest, self).setUp()
        self.username = 'parser_stack_test_user'

        self.ctx = utils.dummy_context()

        self.stack = stack.Stack(self.ctx, 'resolve_test_stack',
                                 template.Template(empty_template))

    def resolve(self, snippet):
        return function.resolve(self.stack.t.parse(self.stack.defn, snippet))

    def test_join_split(self):
        # join
        snippet = {'Fn::Join': [';', ['one', 'two', 'three']]}
        self.assertEqual('one;two;three', self.resolve(snippet))

        # join then split
        snippet = {'Fn::Split': [';', snippet]}
        self.assertEqual(['one', 'two', 'three'], self.resolve(snippet))

    def test_split_join_split_join(self):
        # each snippet in this test encapsulates
        # the snippet from the previous step, leading
        # to increasingly nested function calls

        # split
        snippet = {'Fn::Split': [',', 'one,two,three']}
        self.assertEqual(['one', 'two', 'three'], self.resolve(snippet))

        # split then join
        snippet = {'Fn::Join': [';', snippet]}
        self.assertEqual('one;two;three', self.resolve(snippet))

        # split then join then split
        snippet = {'Fn::Split': [';', snippet]}
        self.assertEqual(['one', 'two', 'three'], self.resolve(snippet))

        # split then join then split then join
        snippet = {'Fn::Join': ['-', snippet]}
        self.assertEqual('one-two-three', self.resolve(snippet))

    def test_join_recursive(self):
        raw = {'Fn::Join': ['\n', [{'Fn::Join':
                                   [' ', ['foo', 'bar']]}, 'baz']]}
        self.assertEqual('foo bar\nbaz', self.resolve(raw))

    def test_join_not_string(self):
        snippet = {'Fn::Join': ['\n', [{'Fn::Join':
                                        [' ', ['foo', 45]]}, 'baz']]}
        error = self.assertRaises(TypeError,
                                  self.resolve, snippet)
        self.assertIn('45', str(error))

    def test_base64_replace(self):
        raw = {'Fn::Base64': {'Fn::Replace': [
            {'foo': 'bar'}, 'Meet at the foo']}}
        self.assertEqual('Meet at the bar', self.resolve(raw))

    def test_replace_base64(self):
        raw = {'Fn::Replace': [{'foo': 'bar'}, {
            'Fn::Base64': 'Meet at the foo'}]}
        self.assertEqual('Meet at the bar', self.resolve(raw))

    def test_nested_selects(self):
        data = {
            'a': ['one', 'two', 'three'],
            'b': ['een', 'twee', {'d': 'D', 'e': 'E'}]
        }
        raw = {'Fn::Select': ['a', data]}
        self.assertEqual(data['a'], self.resolve(raw))

        raw = {'Fn::Select': ['b', data]}
        self.assertEqual(data['b'], self.resolve(raw))

        raw = {
            'Fn::Select': ['1', {
                'Fn::Select': ['b', data]
            }]
        }
        self.assertEqual('twee', self.resolve(raw))

        raw = {
            'Fn::Select': ['e', {
                'Fn::Select': ['2', {
                    'Fn::Select': ['b', data]
                }]
            }]
        }
        self.assertEqual('E', self.resolve(raw))

    def test_member_list_select(self):
        snippet = {'Fn::Select': ['metric', {"Fn::MemberListToMap": [
            'Name', 'Value', ['.member.0.Name=metric',
                              '.member.0.Value=cpu',
                              '.member.1.Name=size',
                              '.member.1.Value=56']]}]}
        self.assertEqual('cpu', self.resolve(snippet))



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import functools

from oslo_utils import strutils

from heat.common.i18n import _
from heat.engine import constraints as constr
from heat.engine import support

from oslo_log import log as logging

LOG = logging.getLogger(__name__)


class Schema(constr.Schema):
    """Simple schema class for attributes.

    Schema objects are serializable to dictionaries following a superset of
    the HOT input Parameter schema using dict().
    """

    KEYS = (
        DESCRIPTION, TYPE
    ) = (
        'description', 'type',
    )

    CACHE_MODES = (
        CACHE_LOCAL,
        CACHE_NONE
    ) = (
        'cache_local',
        'cache_none'
    )

    TYPES = (
        UNKNOWN, STRING, MAP, LIST, INTEGER, BOOLEAN
    ) = (
        None, 'String', 'Map', 'List', 'Integer', 'Boolean'
    )

    def __init__(self, description=None,
                 support_status=support.SupportStatus(),
                 cache_mode=CACHE_LOCAL,
                 type=UNKNOWN):
        super(Schema, self).__init__(type, description)
        self.support_status = support_status
        self.cache_mode = cache_mode
        self.validate()

    def __getitem__(self, key):
        if key == self.DESCRIPTION:
            if self.description is not None:
                return self.description

        elif key == self.TYPE:
            if self.type is not None:
                return self.type.lower()

        raise KeyError(key)

    @classmethod
    def from_attribute(cls, schema_dict):
        """Return a Property Schema corresponding to a Attribute Schema."""
        msg = 'Old attribute schema is not supported'
        assert isinstance(schema_dict, cls), msg
        return schema_dict


def schemata(schema):
    """Return dictionary of Schema objects for given dictionary of schemata."""
    return dict((n, Schema.from_attribute(s)) for n, s in schema.items())


class Attribute(object):
    """An Attribute schema."""

    def __init__(self, attr_name, schema):
        """Initialise with a name and schema.

        :param attr_name: the name of the attribute
        :param schema: attribute schema
        """
        self.name = attr_name
        self.schema = Schema.from_attribute(schema)

    def support_status(self):
        return self.schema.support_status

    def as_output(self, resource_name, template_type='cfn'):
        """Output entry for a provider template with the given resource name.

        :param resource_name: the logical name of the provider resource
        :param template_type: the template type to generate
        :returns: This attribute as a template 'Output' entry for
                  cfn template and 'output' entry for hot template
        """
        if template_type == 'hot':
            return {
                "value": {"get_attr": [resource_name, self.name]},
                "description": self.schema.description
            }
        else:
            return {
                "Value": {"Fn::GetAtt": [resource_name, self.name]},
                "Description": self.schema.description
            }


def _stack_id_output(resource_name, template_type='cfn'):
    if template_type == 'hot':
        return {
            "value": {"get_resource": resource_name},
        }
    else:
        return {
            "Value": {"Ref": resource_name},
        }


BASE_ATTRIBUTES = (SHOW_ATTR, ) = ('show', )

# Returned by function.dep_attrs() to indicate that all attributes are
# referenced
ALL_ATTRIBUTES = '*'


class Attributes(collections.abc.Mapping):
    """Models a collection of Resource Attributes."""

    def __init__(self, res_name, schema, resolver):
        self._resource_name = res_name
        self._resolver = resolver
        self.set_schema(schema)
        self.reset_resolved_values()

        assert ALL_ATTRIBUTES not in schema, \
            "Invalid attribute name '%s'" % ALL_ATTRIBUTES

    def reset_resolved_values(self):
        if hasattr(self, '_resolved_values'):
            self._has_new_resolved = len(self._resolved_values) > 0
        else:
            self._has_new_resolved = False
        self._resolved_values = {}

    def set_schema(self, schema):
        self._attributes = self._make_attributes(schema)

    def get_cache_mode(self, attribute_name):
        """Return the cache mode for the specified attribute.

        If the attribute is not defined in the schema, the default cache
        mode (CACHE_LOCAL) is returned.
        """
        try:
            return self._attributes[attribute_name].schema.cache_mode
        except KeyError:
            return Schema.CACHE_LOCAL

    @staticmethod
    def _make_attributes(schema):
        return dict((n, Attribute(n, d)) for n, d in schema.items())

    @staticmethod
    def as_outputs(resource_name, resource_class, template_type='cfn'):
        """Dict of Output entries for a provider template with resource name.

        :param resource_name: logical name of the resource
        :param resource_class: resource implementation class
        :returns: The attributes of the specified resource_class as a template
                  Output map
        """
        attr_schema = {}
        for name, schema_data in resource_class.attributes_schema.items():
            schema = Schema.from_attribute(schema_data)
            if schema.support_status.status != support.HIDDEN:
                attr_schema[name] = schema
        attr_schema.update(resource_class.base_attributes_schema)
        attribs = Attributes._make_attributes(attr_schema).items()

        outp = dict((n, att.as_output(resource_name,
                                      template_type)) for n, att in attribs)
        outp['OS::stack_id'] = _stack_id_output(resource_name, template_type)
        return outp

    @staticmethod
    def schema_from_outputs(json_snippet):
        if json_snippet:
            return dict((k, Schema(v.get("Description")))
                        for k, v in json_snippet.items())
        return {}

    def _validate_type(self, attrib, value):
        if attrib.schema.type == attrib.schema.STRING:
            if not isinstance(value, str):
                LOG.warning("Attribute %(name)s is not of type "
                            "%(att_type)s",
                            {'name': attrib.name,
                             'att_type': attrib.schema.STRING})
        elif attrib.schema.type == attrib.schema.LIST:
            if (not isinstance(value, collections.abc.Sequence)
                    or isinstance(value, str)):
                LOG.warning("Attribute %(name)s is not of type "
                            "%(att_type)s",
                            {'name': attrib.name,
                             'att_type': attrib.schema.LIST})
        elif attrib.schema.type == attrib.schema.MAP:
            if not isinstance(value, collections.abc.Mapping):
                LOG.warning("Attribute %(name)s is not of type "
                            "%(att_type)s",
                            {'name': attrib.name,
                             'att_type': attrib.schema.MAP})
        elif attrib.schema.type == attrib.schema.INTEGER:
            if not isinstance(value, int):
                LOG.warning("Attribute %(name)s is not of type "
                            "%(att_type)s",
                            {'name': attrib.name,
                             'att_type': attrib.schema.INTEGER})
        elif attrib.schema.type == attrib.schema.BOOLEAN:
            try:
                strutils.bool_from_string(value, strict=True)
            except ValueError:
                LOG.warning("Attribute %(name)s is not of type "
                            "%(att_type)s",
                            {'name': attrib.name,
                             'att_type': attrib.schema.BOOLEAN})

    @property
    def cached_attrs(self):
        return self._resolved_values

    @cached_attrs.setter
    def cached_attrs(self, c_attrs):
        if c_attrs is None:
            self._resolved_values = {}
        else:
            self._resolved_values = c_attrs
        self._has_new_resolved = False

    def set_cached_attr(self, key, value):
        self._resolved_values[key] = value
        self._has_new_resolved = True

    def has_new_cached_attrs(self):
        """Returns True if cached_attrs have changed

        Allows the caller to determine if this instance's cached_attrs
        have been updated since they were initially set (if at all).
        """
        return self._has_new_resolved

    def __getitem__(self, key):
        if key not in self:
            raise KeyError(_('%(resource)s: Invalid attribute %(key)s') %
                           dict(resource=self._resource_name, key=key))

        attrib = self._attributes.get(key)
        if attrib.schema.cache_mode == Schema.CACHE_NONE:
            return self._resolver(key)

        if key in self._resolved_values:
            return self._resolved_values[key]

        value = self._resolver(key)

        if value is not None:
            # validate the value against its type
            self._validate_type(attrib, value)
            # only store if not None, it may resolve to an actual value
            # on subsequent calls
            self.set_cached_attr(key, value)
        return value

    def __len__(self):
        return len(self._attributes)

    def __contains__(self, key):
        return key in self._attributes

    def __iter__(self):
        return iter(self._attributes)

    def __repr__(self):
        return ("Attributes for %s:\n\t" % self._resource_name +
                '\n\t'.join(self.values()))


def select_from_attribute(attribute_value, path):
    """Select an element from an attribute value.

    :param attribute_value: the attribute value.
    :param path: a list of path components to select from the attribute.
    :returns: the selected attribute component value.
    """
    def get_path_component(collection, key):
        if not isinstance(collection, (collections.abc.Mapping,
                                       collections.abc.Sequence)):
            raise TypeError(_("Can't traverse attribute path"))

        if not isinstance(key, (str, int)):
            raise TypeError(_('Path components in attributes must be strings'))

        return collection[key]

    try:
        return functools.reduce(get_path_component, path, attribute_value)
    except (KeyError, IndexError, TypeError):
        return None



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from heat.common import exception
from heat.common import template_format
from heat.engine.resources.openstack.barbican import secret
from heat.engine import rsrc_defn
from heat.engine import scheduler
from heat.tests import common
from heat.tests import utils

stack_template = '''
heat_template_version: 2013-05-23
description: Test template
resources:
  secret:
    type: OS::Barbican::Secret
    properties:
      name: foobar-secret
'''


class FakeSecret(object):

    def __init__(self, name):
        self.name = name

    def store(self):
        return self.name


class TestSecret(common.HeatTestCase):

    def setUp(self):
        super(TestSecret, self).setUp()

        self.patcher_client = mock.patch.object(secret.Secret, 'client')
        mock_client = self.patcher_client.start()
        self.barbican = mock_client.return_value

        self.stack = utils.parse_stack(template_format.parse(stack_template))
        self.stack.validate()
        resource_defns = self.stack.t.resource_definitions(self.stack)
        self.res_template = resource_defns['secret']
        self.res = self._create_resource('foo', self.res_template, self.stack)

    def tearDown(self):
        super(TestSecret, self).tearDown()
        self.patcher_client.stop()

    def _create_resource(self, name, snippet, stack):
        res = secret.Secret(name, snippet, stack)
        self.barbican.secrets.create.return_value = FakeSecret(name + '_id')
        scheduler.TaskRunner(res.create)()
        return res

    def test_create_secret(self):
        expected_state = (self.res.CREATE, self.res.COMPLETE)
        self.assertEqual(expected_state, self.res.state)
        args = self.barbican.secrets.create.call_args[1]
        self.assertEqual('foobar-secret', args['name'])
        self.assertEqual('opaque', args['secret_type'])

    def test_attributes(self):
        mock_secret = mock.Mock()
        mock_secret.status = 'test-status'
        self.barbican.secrets.get.return_value = mock_secret
        mock_secret.payload = 'foo'

        self.assertEqual('test-status', self.res.FnGetAtt('status'))
        self.assertEqual('foo', self.res.FnGetAtt('decrypted_payload'))

    def test_attributes_handles_exceptions(self):
        self.barbican.barbican_client.HTTPClientError = Exception
        self.barbican.secrets.get.side_effect = Exception('boom')
        self.assertRaises(self.barbican.barbican_client.HTTPClientError,
                          self.res.FnGetAtt, 'order_ref')

    def test_create_secret_sets_resource_id(self):
        self.assertEqual('foo_id', self.res.resource_id)

    def test_create_secret_with_plain_text(self):
        content_type = 'text/plain'
        props = {
            'name': 'secret',
            'payload': 'foobar',
            'payload_content_type': content_type,
        }
        defn = rsrc_defn.ResourceDefinition('secret',
                                            'OS::Barbican::Secret',
                                            props)
        res = self._create_resource(defn.name, defn, self.stack)

        args = self.barbican.secrets.create.call_args[1]
        self.assertEqual('foobar', args[res.PAYLOAD])
        self.assertEqual(content_type, args[res.PAYLOAD_CONTENT_TYPE])

    def test_create_secret_with_octet_stream(self):
        content_type = 'application/octet-stream'
        props = {
            'name': 'secret',
            'payload': 'foobar',
            'payload_content_type': content_type,
        }
        defn = rsrc_defn.ResourceDefinition('secret',
                                            'OS::Barbican::Secret',
                                            props)
        res = self._create_resource(defn.name, defn, self.stack)

        args = self.barbican.secrets.create.call_args[1]
        self.assertEqual('foobar', args[res.PAYLOAD])
        self.assertEqual(content_type, args[res.PAYLOAD_CONTENT_TYPE])

    def test_create_secret_other_content_types_not_allowed(self):
        props = {
            'name': 'secret',
            'payload_content_type': 'not/allowed',
        }
        defn = rsrc_defn.ResourceDefinition('secret',
                                            'OS::Barbican::Secret',
                                            props)
        self.assertRaises(exception.ResourceFailure,
                          self._create_resource, defn.name, defn,
                          self.stack)

    def test_validate_content_type_without_payload(self):
        props = {
            'name': 'secret',
            'payload_content_type': 'text/plain',
        }
        defn = rsrc_defn.ResourceDefinition('secret',
                                            'OS::Barbican::Secret',
                                            props)
        res = self._create_resource(defn.name, defn, self.stack)
        msg = "payload_content_type cannot be specified without payload."
        self.assertRaisesRegex(exception.ResourcePropertyDependency,
                               msg, res.validate)

    def test_validate_octet_stream_without_encoding(self):
        props = {
            'name': 'secret',
            'payload': 'foobar',
            'payload_content_type': 'application/octet-stream',
        }
        defn = rsrc_defn.ResourceDefinition('secret',
                                            'OS::Barbican::Secret',
                                            props)
        res = self._create_resource(defn.name, defn, self.stack)
        msg = ("Property unspecified. For 'application/octet-stream' value of "
               "'payload_content_type' property, 'payload_content_encoding' "
               "property must be specified.")
        self.assertRaisesRegex(exception.StackValidationFailed,
                               msg, res.validate)

    def test_validate_base64(self):
        props = {
            'name': 'secret',
            'payload': 'foobar',
            'payload_content_type': 'application/octet-stream',
            'payload_content_encoding': 'base64'
        }
        defn = rsrc_defn.ResourceDefinition('secret',
                                            'OS::Barbican::Secret',
                                            props)
        res = self._create_resource(defn.name, defn, self.stack)
        msg = ("Invalid payload for specified 'base64' value of "
               "'payload_content_encoding' property.")
        self.assertRaisesRegex(exception.StackValidationFailed,
                               msg, res.validate)

    def test_validate_encoding_dependency(self):
        props = {
            'name': 'secret',
            'payload': 'foobar',
            'payload_content_type': 'text/plain',
            'payload_content_encoding': 'base64'
        }
        defn = rsrc_defn.ResourceDefinition('secret',
                                            'OS::Barbican::Secret',
                                            props)
        res = self._create_resource(defn.name, defn, self.stack)
        msg = ("payload_content_encoding property should only be specified "
               "for payload_content_type with value "
               "application/octet-stream.")
        self.assertRaisesRegex(exception.ResourcePropertyValueDependency,
                               msg, res.validate)



# Copyright 2014 Intel Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from oslo_serialization import jsonutils as json
from oslo_versionedobjects import fields


class Json(fields.FieldType):
    def coerce(self, obj, attr, value):
        if isinstance(value, str):
            loaded = json.loads(value)
            return loaded
        return value

    def from_primitive(self, obj, attr, value):
        return self.coerce(obj, attr, value)

    def to_primitive(self, obj, attr, value):
        return json.dumps(value)


class JsonField(fields.AutoTypedField):
    AUTO_TYPE = Json()


class ListField(fields.AutoTypedField):
    AUTO_TYPE = fields.List(fields.FieldType())



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from rally import consts
from rally.plugins.openstack import scenario
from rally.plugins.openstack.scenarios.heat import utils
from rally.task import atomic
from rally.task import types
from rally.task import validation


class CustomHeatBenchmark(utils.HeatScenario):
    @atomic.action_timer("heat.show_output_new")
    def _stack_show_output_new(self, stack, output_key):
        """Execute output_show for specified 'output_key'.

        This method uses new output API call.

        :param stack: stack with output_key output.
        :param output_key: The name of the output.
        """
        self.clients("heat").stacks.output_show(stack.id, output_key)

    @atomic.action_timer("heat.show_output_old")
    def _stack_show_output_old(self, stack, output_key):
        """Execute output_show for specified 'output_key'.

        This method uses old way for getting output value.
        It gets whole stack object and then finds necessary 'output_key'.

        :param stack: stack with output_key output.
        :param output_key: The name of the output.
        """
        # this code copy-pasted and adopted for rally from old client version
        # https://github.com/openstack/python-heatclient/blob/0.8.0/heatclient/
        # v1/shell.py#L682-L699
        stack = self.clients("heat").stacks.get(stack_id=stack.id)
        for output in stack.to_dict().get('outputs', []):
            if output['output_key'] == output_key:
                break

    @atomic.action_timer("heat.list_output_new")
    def _stack_list_output_new(self, stack):
        """Execute output_list for specified 'stack'.

        This method uses new output API call.

        :param stack: stack to call output-list.
        """
        self.clients("heat").stacks.output_list(stack.id)

    @atomic.action_timer("heat.list_output_old")
    def _stack_list_output_old(self, stack):
        """Execute output_list for specified 'stack'.

        This method uses old way for getting output value.
        It gets whole stack object and then prints all outputs
        belongs this stack.

        :param stack: stack to call output-list.
        """
        # this code copy-pasted and adopted for rally from old client version
        # https://github.com/openstack/python-heatclient/blob/0.8.0/heatclient/
        # v1/shell.py#L649-L663
        stack = self.clients("heat").stacks.get(stack_id=stack.id)
        stack.to_dict()['outputs']

    @types.set(template_path=types.FileType, files=types.FileTypeDict)
    @validation.required_services(consts.Service.HEAT)
    @validation.required_openstack(users=True)
    @scenario.configure(context={"cleanup": ["heat"]})
    def create_stack_and_show_output_old(self, template_path, output_key,
                                         parameters=None, files=None,
                                         environment=None):
        """Create stack and show output by using old algorithm.

        Measure performance of the following commands:
        heat stack-create
        heat output-show
        heat stack-delete

        :param template_path: path to stack template file
        :param parameters: parameters to use in heat template
        :param files: files used in template
        :param environment: stack environment definition
        """
        stack = self._create_stack(
            template_path, parameters, files, environment)
        self._stack_show_output_old(stack, output_key)

    @types.set(template_path=types.FileType, files=types.FileTypeDict)
    @validation.required_services(consts.Service.HEAT)
    @validation.required_openstack(users=True)
    @scenario.configure(context={"cleanup": ["heat"]})
    def create_stack_and_show_output_new(self, template_path, output_key,
                                         parameters=None, files=None,
                                         environment=None):
        """Create stack and show output by using new algorithm.

        Measure performance of the following commands:
        heat stack-create
        heat output-show
        heat stack-delete

        :param template_path: path to stack template file
        :param output_key: the stack output key that corresponds to
                           the scaling webhook
        :param parameters: parameters to use in heat template
        :param files: files used in template
        :param environment: stack environment definition
        """
        stack = self._create_stack(
            template_path, parameters, files, environment)
        self._stack_show_output_new(stack, output_key)

    @types.set(template_path=types.FileType, files=types.FileTypeDict)
    @validation.required_services(consts.Service.HEAT)
    @validation.required_openstack(users=True)
    @scenario.configure(context={"cleanup": ["heat"]})
    def create_stack_and_list_output_old(self, template_path,
                                         parameters=None, files=None,
                                         environment=None):
        """Create stack and list outputs by using old algorithm.

        Measure performance of the following commands:
        heat stack-create
        heat output-list
        heat stack-delete

        :param template_path: path to stack template file
        :param parameters: parameters to use in heat template
        :param files: files used in template
        :param environment: stack environment definition
        """
        stack = self._create_stack(
            template_path, parameters, files, environment)
        self._stack_list_output_old(stack)

    @types.set(template_path=types.FileType, files=types.FileTypeDict)
    @validation.required_services(consts.Service.HEAT)
    @validation.required_openstack(users=True)
    @scenario.configure(context={"cleanup": ["heat"]})
    def create_stack_and_list_output_new(self, template_path,
                                         parameters=None, files=None,
                                         environment=None):
        """Create stack and list outputs by using new algorithm.

        Measure performance of the following commands:
        heat stack-create
        heat output-list
        heat stack-delete

        :param template_path: path to stack template file
        :param output_key: the stack output key that corresponds to
                           the scaling webhook
        :param parameters: parameters to use in heat template
        :param files: files used in template
        :param environment: stack environment definition
        """
        stack = self._create_stack(
            template_path, parameters, files, environment)
        self._stack_list_output_new(stack)
